
# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import


class Queue(object):
    """
    A SET WITH ADDED ORDER MAINTAINED

    +------------+---------+----------+
    | Uniqueness | Ordered | Type     |
    +------------+---------+----------+
    |     Yes    |   Yes   | Queue    |
    |     Yes    |   No    | Set      |
    |     No     |   Yes   | List     |
    |     No     |   No    | Multiset |
    +------------+---------+----------+
    """
    def __init__(self):
        self.set = set()
        self.list = []

    def __nonzero__(self):
        return len(self.list) > 0

    def __len__(self):
        return self.list.__len__()

    def add(self, value):
        if value in self.set:
            return self
        self.set.add(value)
        self.list.append(value)

    def push(self, value):
        self.add(value)

    def extend(self, values):
        for v in values:
            self.add(v)

    def pop(self):
        if len(self.list) == 0:
            return None

        output = self.list.pop(0)
        self.set.remove(output)
        return output




# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import



class Multiset(object):
    """
    Multiset IS ONE MEMBER IN A FAMILY OF USEFUL CONTAINERS

    +------------+---------+----------+
    | Uniqueness | Ordered | Type     |
    +------------+---------+----------+
    |     Yes    |   Yes   | Queue    |
    |     Yes    |   No    | Set      |
    |     No     |   Yes   | List     |
    |     No     |   No    | Multiset |
    +------------+---------+----------+
    """
    def __new__(cls, list=None, key_field=None, count_field=None, allow_negative=False):
        try:
            if allow_negative:
                return _NegMultiset(list, key_field, count_field)
            else:
                return _Multiset(list, key_field, count_field)
        except Exception as e:
            from mo_logs import Log

            Log.error("Not expected", e)

    def add(self, value):
        raise NotImplementedError

    def extend(self, values):
        raise NotImplementedError

    def remove(self, value):
        raise NotImplementedError


class _Multiset(Multiset):

    def __new__(cls, *args):
        return object.__new__(cls, *args)


    def __init__(self, list=None, key_field=None, count_field=None, **kwargs):
        if not key_field and not count_field:
            self.dic = dict()
            if list:
                for i in list:
                    self.add(i)
            return
        else:
            self.dic = {i[key_field]: i[count_field] for i in list}


    def __iter__(self):
        for k, m in self.dic.items():
            for i in range(m):
                yield k


    def items(self):
        return self.dic.items()

    def keys(self):
        return self.dic.keys()

    def add(self, value):
        if value in self.dic:
            self.dic[value] += 1
        else:
            self.dic[value] = 1
        return self

    def extend(self, values):
        for v in values:
            self.add(v)

    def remove(self, value):
        if value not in self.dic:
            from mo_logs import Log

            Log.error("{{value}} is not in multiset",  value= value)
        self._remove(value)

    def copy(self):
        output = _Multiset()
        output.dic = self.dic.copy()
        return output


    def _remove(self, value):
        count = self.dic.get(value)
        if count == None:
            return

        count -= 1
        if count == 0:
            del (self.dic[value])
        else:
            self.dic[value] = count

    def __sub__(self, other):
        output = self.copy()
        for o in other:
            output._remove(o)
        return output

    def __add__(self, other):
        output = self.copy()

        if isinstance(other, Multiset):
            for k, c in other.dic.items():
                output.dic[k] = output.dic.get(k, 0) + c
        else:
            for o in other:
                output.add(o)
        return output

    def __set__(self, other):
        return set(self.dic.keys())

    def __len__(self):
        return sum(self.dic.values())

    def __nonzero__(self):
        if self.dic:
            return True
        return False


    def count(self, value):
        if value in self.dic:
            return self.dic[value]
        else:
            return 0


class _NegMultiset(Multiset):
    def __new__(cls, *args, **kwargs):
            return object.__new__(cls, *args, **kwargs)

    def __init__(self, list=None, key_field=None, count_field=None, **kwargs):
        if not key_field and not count_field:
            self.dic = dict()
            if list:
                for i in list:
                    self.add(i)
            return
        else:
            self.dic = {i[key_field]: i[count_field] for i in list}


    # def __iter__(self):
    #     for k, m in self.dic.items():
    #         for i in range(m):
    #             yield k


    def items(self):
        return self.dic.items()

    def keys(self):
        return self.dic.keys()

    def add(self, value, amount=None):
        count = self.dic.get(value)
        if amount == None:
            amount = 1
        elif amount == 0:
            return self

        if not count:
            self.dic[value] = amount
        elif count == -amount:
            del (self.dic[value])
        else:
            self.dic[value] = count + amount

        return self

    def extend(self, values):
        for v in values:
            self.add(v)

    def remove(self, value):
        return self.add(value, -1)


    def copy(self):
        output = _NegMultiset()
        output.dic = self.dic.copy()
        return output


    def __add__(self, other):
        output = self.copy()

        if isinstance(other, _NegMultiset):
            for k, c in other.dic.items():
                output.add(k, c)
        else:
            for o in other:
                output.add(o)

        return output

    def __sub__(self, other):
        if not other:
            return self

        output = self.copy()
        for o in other:
            output.remove(o)
        return output

    def __set__(self, other):
        return set(self.dic.keys())

    def __len__(self):
        return sum(self.dic.values())

    def __nonzero__(self):
        if self.dic:
            return True
        return False


    def count(self, value):
        if value in self.dic:
            return self.dic[value]
        else:
            return 0



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals


from mo_collections.unique_index import UniqueIndex

def reverse(values):
    """
    REVERSE - WITH NO SIDE EFFECTS!
    """
    output = list(values)
    output.reverse()
    return output





# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import mo_json
from mo_dots import Data, wrap
from mo_files import File
from mo_logs import Log
from mo_logs.exceptions import suppress_exception
from mo_math.randoms import Random
from mo_threads import Lock, Signal, THREAD_STOP

DEBUG = True


class PersistentQueue(object):
    """
    THREAD-SAFE, PERSISTENT QUEUE

    CAN HANDLE MANY PRODUCERS, BUT THE pop(), commit() IDIOM CAN HANDLE ONLY
    ONE CONSUMER.

    IT IS IMPORTANT YOU commit() or close(), OTHERWISE NOTHING COMES OFF THE QUEUE
    """

    def __init__(self, _file):
        """
        file - USES FILE FOR PERSISTENCE
        """
        self.file = File.new_instance(_file)
        self.lock = Lock("lock for persistent queue using file " + self.file.name)
        self.please_stop = Signal()
        self.db = Data()
        self.pending = []

        if self.file.exists:
            for line in self.file:
                with suppress_exception:
                    delta = mo_json.json2value(line)
                    apply_delta(self.db, delta)
            if self.db.status.start == None:  # HAPPENS WHEN ONLY ADDED TO QUEUE, THEN CRASH
                self.db.status.start = 0
            self.start = self.db.status.start

            # SCRUB LOST VALUES
            lost = 0
            for k in self.db.keys():
                with suppress_exception:
                    if k!="status" and int(k) < self.start:
                        self.db[k] = None
                        lost += 1
                  # HAPPENS FOR self.db.status, BUT MAYBE OTHER PROPERTIES TOO
            if lost:
                Log.warning("queue file had {{num}} items lost",  num= lost)

            if DEBUG:
                Log.note("Persistent queue {{name}} found with {{num}} items", name=self.file.abspath, num=len(self))
        else:
            self.db.status = Data(
                start=0,
                end=0
            )
            self.start = self.db.status.start
            if DEBUG:
                Log.note("New persistent queue {{name}}", name=self.file.abspath)

    def _add_pending(self, delta):
        delta = wrap(delta)
        self.pending.append(delta)

    def _apply_pending(self):
        for delta in self.pending:
            apply_delta(self.db, delta)
        self.pending = []


    def __iter__(self):
        """
        BLOCKING ITERATOR
        """
        while not self.please_stop:
            try:
                value = self.pop()
                if value is not THREAD_STOP:
                    yield value
            except Exception as e:
                Log.warning("Tell me about what happened here", cause=e)
        if DEBUG:
            Log.note("queue iterator is done")

    def add(self, value):
        with self.lock:
            if self.closed:
                Log.error("Queue is closed")

            if value is THREAD_STOP:
                if DEBUG:
                    Log.note("Stop is seen in persistent queue")
                self.please_stop.go()
                return

            self._add_pending({"add": {str(self.db.status.end): value}})
            self.db.status.end += 1
            self._add_pending({"add": {"status.end": self.db.status.end}})
            self._commit()
        return self

    def __len__(self):
        with self.lock:
            return self.db.status.end - self.start

    def __getitem__(self, item):
        return self.db[str(item + self.start)]

    def pop(self, timeout=None):
        """
        :param timeout: OPTIONAL DURATION
        :return: None, IF timeout PASSES
        """
        with self.lock:
            while not self.please_stop:
                if self.db.status.end > self.start:
                    value = self.db[str(self.start)]
                    self.start += 1
                    return value

                if timeout is not None:
                    with suppress_exception:
                        self.lock.wait(timeout=timeout)
                        if self.db.status.end <= self.start:
                            return None
                else:
                    self.lock.wait()

            if DEBUG:
                Log.note("persistent queue already stopped")
            return THREAD_STOP

    def pop_all(self):
        """
        NON-BLOCKING POP ALL IN QUEUE, IF ANY
        """
        with self.lock:
            if self.please_stop:
                return [THREAD_STOP]
            if self.db.status.end == self.start:
                return []

            output = []
            for i in range(self.start, self.db.status.end):
                output.append(self.db[str(i)])

            self.start = self.db.status.end
            return output

    def rollback(self):
        with self.lock:
            if self.closed:
                return
            self.start = self.db.status.start
            self.pending = []

    def commit(self):
        with self.lock:
            if self.closed:
                Log.error("Queue is closed, commit not allowed")

            try:
                self._add_pending({"add": {"status.start": self.start}})
                for i in range(self.db.status.start, self.start):
                    self._add_pending({"remove": str(i)})

                if self.db.status.end - self.start < 10 or Random.range(0, 1000) == 0:  # FORCE RE-WRITE TO LIMIT FILE SIZE
                    # SIMPLY RE-WRITE FILE
                    if DEBUG:
                        Log.note("Re-write {{num_keys}} keys to persistent queue", num_keys=self.db.status.end - self.start)

                        for k in self.db.keys():
                            if k == "status" or int(k) >= self.db.status.start:
                                continue
                            Log.error("Not expecting {{key}}", key=k)
                    self._commit()
                    self.file.write(mo_json.value2json({"add": self.db}) + "\n")
                else:
                    self._commit()
            except Exception as e:
                raise e

    def _commit(self):
        self.file.append("\n".join(mo_json.value2json(p) for p in self.pending))
        self._apply_pending()

    def close(self):
        self.please_stop.go()
        with self.lock:
            if self.db is None:
                return

            self.add(THREAD_STOP)

            if self.db.status.end == self.start:
                if DEBUG:
                    Log.note("persistent queue clear and closed")
                self.file.delete()
            else:
                if DEBUG:
                    Log.note("persistent queue closed with {{num}} items left", num=len(self))
                try:
                    self._add_pending({"add": {"status.start": self.start}})
                    for i in range(self.db.status.start, self.start):
                        self._add_pending({"remove": str(i)})
                    self.file.write(mo_json.value2json({"add": self.db}) + "\n" + ("\n".join(mo_json.value2json(p) for p in self.pending)) + "\n")
                    self._apply_pending()
                except Exception as e:
                    raise e
            self.db = None

    @property
    def closed(self):
        with self.lock:
            return self.db is None


def apply_delta(value, delta):
    if delta.add:
        for k, v in delta.add.items():
            value[k] = v
    elif delta.remove:
        value[delta.remove] = None



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from __future__ import unicode_literals

from mo_testing.fuzzytestcase import FuzzyTestCase

from mo_collections.matrix import index_to_coordinate


class TestUniqueIndex(FuzzyTestCase):

    def test_zero_dim(self):
        f = index_to_coordinate([])

        self.assertAlmostEqual(f(-1), tuple())
        self.assertAlmostEqual(f(0), tuple())
        self.assertAlmostEqual(f(1), tuple())
        self.assertAlmostEqual(f(2), tuple())
        self.assertAlmostEqual(f(3), tuple())

    def test_one_dim(self):
        f = index_to_coordinate([2])

        # self.assertAlmostEqual(f(-1), 1,)
        self.assertAlmostEqual(f(0), (0,))
        self.assertAlmostEqual(f(1), (1,))
        # self.assertAlmostEqual(f(2), (0,))
        # self.assertAlmostEqual(f(3), (1,))

    def test_two_dim(self):
        f = index_to_coordinate([2, 3])

        # self.assertAlmostEqual(f(-1), (1, 2))
        self.assertAlmostEqual(f(0), (0,0))
        self.assertAlmostEqual(f(1), (0,1))
        self.assertAlmostEqual(f(2), (0,2))
        self.assertAlmostEqual(f(3), (1,0))
        self.assertAlmostEqual(f(4), (1,1))
        self.assertAlmostEqual(f(5), (1,2))
        # self.assertAlmostEqual(f(6), (0,0))
        # self.assertAlmostEqual(f(7), (0,1))
        # self.assertAlmostEqual(f(8), (0,2))




# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from distutils.util import convert_path
import os
from setuptools import setup


root = os.path.abspath(os.path.dirname(__file__))
path = lambda *p: os.path.join(root, *p)
try:
    long_desc = open(path('README.txt')).read()
except Exception:
    long_desc = "<Missing README.txt>"
    print("Missing README.txt")


def find_packages(where='.', lib_prefix='', exclude=()):
    """
    SNAGGED FROM distribute-0.6.49-py2.7.egg/setuptools/__init__.py
    """
    out = []
    stack=[(convert_path(where), lib_prefix)]
    while stack:
        where,prefix = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where,name)
            if ('.' not in name and os.path.isdir(fn) and
                os.path.isfile(os.path.join(fn,'__init__.py'))
            ):
                out.append(prefix+name); stack.append((fn,prefix+name+'.'))
    for pat in list(exclude)+['ez_setup', 'distribute_setup']:
        from fnmatch import fnmatchcase
        out = [item for item in out if not fnmatchcase(item,pat)]
    return out




setup(
    name='mo-collections',
    version="1.2.18029",
    description='More Collections! Some useful data structures for dealing with Data',
    long_description=long_desc,
    author='Kyle Lahnakoski',
    author_email='kyle@lahnakoski.com',
    url='https://github.com/klahnakoski/mo-collections',
    license='MPL 2.0',
    packages=find_packages(".", lib_prefix=""),
    install_requires=["mo-dots","mo-future","mo-kwargs","mo-logs"],
    include_package_data=True,
    zip_safe=False,
    classifiers=[  #https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
    ]
)
